---------------------------------------------------------------------------
-- Core module of CLOPWH. <br />
-- Author : combacsa (combacsa@gmail.com, aka Byun, Kyuhong aka Francisco)
---------------------------------------------------------------------------

local require, pairs, ipairs, table, setmetatable, getmetatable, type, tonumber, tostring = require, pairs, ipairs, table, setmetatable, getmetatable, type, tonumber, tostring
local util = require "clopwh.util"

module("clopwh.core")


---------------------------------------------------------------------------
-- Stores all metatables which describe each Class.
-- @field class_id metatable for corresponding Class
local class_array = {}

---------------------------------------------------------------------------
-- Matches Class and its corresponding class_id
-- @field Class class_id of given Class
local class_index = {}


--------------------------------------------------------------------------------
-- Return a function that matches with <code>key</code> for given <code>class_id
-- </code> and its ancestors.
-- @param class_id Unique id for class.
-- @param key Given function name to lookup.
-- @return The function matching with key. If nothing exist, nil.
local function lookup (class_id, key)
	for _, class_id in ipairs(class_array[class_id].ancestors) do
		declaration = class_array[class_id].declaration
		if declaration[key] ~= nil then
			return declaration[key]
		end
	end
	return nil
end

--------------------------------------------------------------------------------
-- Return a new Object for given class with calling its constructor with given
-- parameters. If there is no constructor, seek to its ancestors.
-- @param class_id Unique id for given class
-- @param ...      Parameters passed to the class constructor
-- @return         Object created by that constructor
local function constructor (class_id,  ...)
	--   TODO : 다중 상속을 했는데 생성자가 없으면 어째야 좋을까?
	local o = {}

	setmetatable(o, class_array[class_id])
	local initialize = lookup(class_id, "initialize")
	if initialize then
		initialize(o, ...)
	end
	return o
end

--------------------------------------------------------------------------------
-- Convert array of Class into array of <code>class_id.</code>
-- @param class_names table(array)of class names
-- @return table(array)of class ids
local function class_names_to_class_ids(class_names)
        local result = {}
	for i, name in ipairs(class_names) do
		result[i] = class_index[name]
	end
	return result
end

--------------------------------------------------------------------------------
-- Linearize multiple inheritance information.
-- @param class_id       Class' unique id
-- @param direct_parents Class' direct parents
-- @return linearized class inheritance
local function linearize(class_id, direct_parents)
	local class_ids = class_names_to_class_ids(direct_parents)
	local list_of_list = {}
	for i, item in ipairs(class_ids) do
		list_of_list[#list_of_list + 1] = class_array[item].ancestors
	end

	return util.merge(class_id, list_of_list)
end

--------------------------------------------------------------------------------
-- Create new lookup function for given class.
-- @param class_id Unique id for given class
-- @return Lookup function which corresponds only with the given class
local function lookup_generator(class_id)
	return function(self, key)
		return function( ... )
			local target_function = lookup(class_id, key)
			return target_function(self, ...)
		end
	end
end

--------------------------------------------------------------------------------
-- Create unary operator overloading function for given class and operator.
-- @param class_id Class id
-- @param operator operator to be overloaded
-- @return lookup function for the class with given operator
local function unary_operator_generator(class_id, operator)
	return function(self)
		local target_function = lookup(class_id, operator)
		return target_function(self)
	end
end

--------------------------------------------------------------------------------
-- Create binary operator overloading function for given class and operator.
-- @param class_id Class id
-- @param operator operator to be overloaded
-- @return lookup function for the class with given operator
local function binary_operator_generator(class_id, operator)
	return function(self, other)
		local target_function = lookup(class_id, operator)
		return target_function(self, other)
	end
end

--------------------------------------------------------------------------------
-- Register new class into class_array and create new Class according to the
-- declaration.
-- @param declaration table which contains Class declaration
-- @param class_function (optional) constructor function for given class
-- @param inherits (optional) inheritance for given class
-- @return Constructor function which is capable of creating object for given class
local function register_class(declaration, class_function, inherits)
	local class_id = #class_array + 1
	local mt = {}

	local class_function = class_function or function ( ... )
		return constructor(class_id, ...)
	end

	local direct_parents = inherits or declaration.inherits or {Object}
	local linearized_ancestors = linearize(class_id, direct_parents)

	if not linearized_ancestors then
		-- C3 Linealization failure.
		-- Rolling back everything.
		return nil -- TODO: error 발생시키기
	end

	mt.__index = lookup_generator(class_id)
	mt.declaration = declaration
	mt.class_function = class_function
	mt.direct_parents = direct_parents
	mt.ancestors = linearized_ancestors
	mt.class_id = class_id

	mt.__add = binary_operator_generator(class_id, "+")
	mt.__mul = binary_operator_generator(class_id, "*")
	mt.__sub = binary_operator_generator(class_id, "-")
	mt.__div = binary_operator_generator(class_id, "/")
	mt.__mod = binary_operator_generator(class_id, "%")
	mt.__pow = binary_operator_generator(class_id, "^")
	mt.__concat = binary_operator_generator(class_id, "..")
	mt.__lt = binary_operator_generator(class_id, "<")
	mt.__le = binary_operator_generator(class_id, "<=")
	mt.__unm = unary_operator_generator(class_id, "(-)")

	if #direct_parents > 0 then -- Special cases
		if (lookup(linearized_ancestors[2], "==") or declaration["=="]) then
			mt.__eq = binary_operator_generator(class_id, "==")
		end
		if (lookup(linearized_ancestors[2], "tostring") or declaration["tostring"]) then
			mt.__tostring = unary_operator_generator(class_id, "tostring")
		end
	end

	class_array[class_id] = mt
	class_index[class_function] = class_id

	return class_function
end

local type_table = {}

--------------------------------------------------------------------------------
-- Returns the type (Class) of given object.
-- @param object Object to know the class
-- @return Class for the object
function get_type(object)
	local type_str = type(object)
	if type_str == "table" then
		local mt = getmetatable(object)
		if mt then
			return mt.class_function or Table
		else
			return Table -- Evaluation 자체가 실행될 때 되겠지 이렇게 해도 ??
		end
	elseif type_str == "function" then
		local result = class_index[object]
		if result then -- Class
			return Class
		else
			return Function
		end
	else
		return type_table[type_str]
	end
end

--------------------------------------------------------------------------------
-- Returns whether given object belongs to the given class or not.
-- @param object     Object to test whether it belongs to given Class
-- @param class_name Class (function) to test whether given object belongs to it
-- @return whether given objec is of given class (according to inheritance)
function is_a(object, class_name)
	object_type = get_type(object)
	object_id = class_index[object_type]
	class_id = class_index[class_name]

	return util.contain_from(class_id, class_array[object_id].ancestors) ~= nil
end

--------------------------------------------------------------------------------
-- Register default Classes.
local function register_default_classes()
	Object = register_class({}, nil, {})
	Class  = register_class({}, function(declaration) return register_class(declaration) end, nil)

	Number = register_class({}, function(x) return tonumber(x) end, nil)
	String = register_class({}, function(x) return tostring(x) end, nil)
	Boolean = register_class({}, function(x) return not not x end, nil)
	Table = register_class({}, function(x) if type(x) == "table" then return util.copy_table(x) else return nil --[[TODO: Error throwing]] end end, nil)
	Nil = register_class({}, function() return nil --[[TODO:Error throwing?]] end, nil)
	Function = register_class({}, function(x) if type(x) == "function" then return x else return nil --[[TODO: Error throwing]] end end, nil)
	Thread = register_class({}, function(x) if type(x) == "thread" then return x else return nil --[[TODO: Error throwing?]] end end, nil)
	Userdata = register_class({}, function(x) if type(x) == "userdata" then return x else return nil --[[TODO: Error throwing?]] end end, nil)

	type_table = {number=Number, string=String, boolean=Boolean, ["nil"]=Nil, ["function"]=Function, ["thread"]=Thread, ["userdata"]=Userdata}
end

-- Run the basic class registeration
register_default_classes()
