---------------------------------------------------------------------------
-- General utilities.
-- Author : combacsa (combacsa@gmail.com, aka Byun, Kyuhong aka Francisco)
---------------------------------------------------------------------------

local ipairs, pairs, table = ipairs, pairs, table

---------------------------------------------------------------------------
-- Module with several utilities that could not fit in a specific module

module "clopwh.util"

---------------------------------------------------------------------------
-- Convert given array into a table. Table contains all elements in given
-- array as Key, and all values associated with any Keys are simply "true".
-- @param array table(array)
-- @return table, whose key is values in given array. All values are true.
function array_to_table(array)
	local result = {}
	for idx, item in ipairs(array) do
		result[item] = true
	end
	return result
end

---------------------------------------------------------------------------
-- Clone given table.
-- @param table table to be copied
-- @return copied table
function copy_table(table)
	local result = {}
	for key in pairs(table) do
		result[key] = table[key]
	end
	return result
end

---------------------------------------------------------------------------
-- Extract only non-empty tables(arrays) from the table(array) of tables.
-- @param seqs table(array) of tables(arrays)
-- @return table(array) of tables(arrays) which empty arrays are excluded
function get_nonempty_seqs(seqs)
	local result = {}
	for seq in pairs(seqs) do
		if #seqs[seq] > 0 then
			result[#result + 1] = seqs[seq]
		end
	end
	return result
end

---------------------------------------------------------------------------
-- Check whether given item is included in the part of given table(array)
-- begins with idx-th index.
-- @param item  item to examine
-- @param table table to examine
-- @param idx   index to start examine
-- @return index of table which given item first occurs. Otherwise nil.
function contain_from(item, table, idx)
	local idx = idx or 1
	for i = idx, #table do
		if table[i] == item then
			return i
		end
	end
	return nil -- Not found
end

---------------------------------------------------------------------------
-- Clone given "table(array) of tables(arrays)".
-- @param seqs table(array) of tables(arrays) to be clonned
-- @return clonned table(array) of tables(arrays)
-- @see util.copy_table
function copy_table_of_tables(seqs)
	local seqs = copy_table(seqs)
	for key in pairs(seqs) do
		seqs[key] = copy_table(seqs[key])
	end
	return seqs
end

--------------------------------------------------------------------------------
-- do C3 Linearization to get proper Method Resolution Order.
-- @param class_id current class' id
-- @param seqs     table(array) of table(array) contains all inheritance
-- @return linearized inheritance table(array)
function merge(class_id, seqs)
	local seqs = copy_table_of_tables(seqs)

	local result = {class_id}
	while true do
		seqs = get_nonempty_seqs(seqs)
		if #seqs == 0 then
			return result
		end
		for idx, seq in ipairs(seqs) do
			cand = seq[1]
			for id, other_seq in ipairs(seqs) do
				if contain_from(cand, other_seq, 2) then
					cand = nil
					break
				end
			end
			if cand then
				break
			end
		end
		if not cand then
			-- Inheritance inconsistency
			return nil
		end
		result[#result + 1] = cand
		for idx, seq in ipairs(seqs) do
			if seq[1] == cand then
				table.remove(seq, 1)
			end
		end
	end
end
